/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wBoundary.h"
#include "wFace.h"
#include "wResults.h"
#include "wGroup.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <unordered_set>
#include <iomanip>
#include <fstream>
using namespace tbox;
using namespace flow;

Boundary::Boundary(std::shared_ptr<MshData> _msh, std::vector<int> const &nos) : Groups(_msh, nos), Cl(0), Cd(0), Cs(0), Cm(0)
{
    // Sanity check
    if (groups.size() != 2)
    {
        std::stringstream err;
        err << "flow::Boundary should be built with exactly 2 physical groups but " << groups.size() << " were given!\n";
        throw std::runtime_error(err.str());
    }
    create();
}

Boundary::Boundary(std::shared_ptr<MshData> _msh, std::vector<std::string> const &names) : Groups(_msh, names), Cl(0), Cd(0), Cs(0), Cm(0)
{
    // Sanity check
    if (groups.size() != 2)
    {
        std::stringstream err;
        err << "flow::Boundary should be built with exactly 2 physical groups but " << groups.size() << " were given!\n";
        throw std::runtime_error(err.str());
    }
    create();
}

/**
 * @brief Create data structure
 * @authors Adrien Crovato
 */
void Boundary::create()
{
    // Get the nodes of Boundary
    for (auto e : groups[0]->tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);
    std::sort(nodes.begin(), nodes.end());
    auto iterator = std::unique(nodes.begin(), nodes.end());
    nodes.resize(std::distance(nodes.begin(), iterator));
    // Create a map between the boundary (surface) nodes and their local index
    for (size_t i = 0; i < nodes.size(); ++i)
        nMap[nodes[i]] = i;

    // Create unordered set of surface faces
    std::unordered_set<Face *, EquFace, EquFace> bndFaces;
    for (auto e : groups[0]->tag->elems)
    {
        Face *f = new Face(e->nodes);
        f->el0 = e;
        bndFaces.insert(f);
    }
    // Find volume faces that are surface faces (valid for TRI3 and TETRA4 only)
    for (auto e : groups[1]->tag->elems)
    {
        size_t idx = 0;
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            // build volume face
            std::vector<Node *> fN(e->nodes.size() - 1);
            for (size_t j = 0; j < e->nodes.size() - 1; ++j)
                fN[j] = e->nodes[(idx + j) % e->nodes.size()];
            Face f(fN);
            // check if face is a surface face
            // for some meshes, 1 tetra4 can have 2 faces that are surface tri3!
            auto it = bndFaces.find(&f);
            if (it != bndFaces.end())
                (*it)->el1 = e;
            idx++;
        }
    }
    // Create a map between the boundary (surface) and volume elements
    for (auto f : bndFaces)
    {
        if (f->el1 == NULL)
        {
            std::stringstream err;
            err << "flow::Boundary: no volume element from tag " << *groups[1]->tag << "could be associated to surface element " << *(f->el0) << "!\n";
            throw std::runtime_error(err.str());
        }
        svMap[f->el0] = f->el1;
        delete f;
    }

    // Resize loads vector
    nLoads.resize(nodes.size(), Eigen::Vector3d::Zero());
}

/**
 * @brief Save nodal data for further post-processing
 * @authors Adrien Crovato
 */
void Boundary::save(std::string const &name, Results const &res)
{
    // Write to file
    std::cout << "writing file: " << name + ".dat"
              << "... " << std::flush;
    std::ofstream outfile;
    outfile.open(name + ".dat");
    // Header
    outfile << "$Boundary data" << std::endl;
    // Aerodynamic coefficients
    outfile << "$Aerodynamic coefficients" << std::endl;
    outfile << "!" << std::fixed
            << std::setw(14) << "Cl"
            << std::setw(15) << "Cd"
            << std::setw(15) << "Cs"
            << std::setw(15) << "Cm"
            << std::endl;
    outfile << std::fixed
            << std::setw(15) << Cl
            << std::setw(15) << Cd
            << std::setw(15) << Cs
            << std::setw(15) << Cm
            << std::endl;
    // Elements (connectvity)
    outfile << "$Elements" << std::endl;
    outfile << groups[0]->tag->elems.size() << std::endl;
    for (auto e : groups[0]->tag->elems)
    {
        outfile << std::fixed
                << std::setw(10) << e->no;
        for (auto n : e->nodes)
            outfile << std::setw(10) << n->no;
        outfile << std::endl;
    }
    //Nodes (data)
    outfile << "$Nodal data" << std::endl;
    outfile << nodes.size() << std::endl;
    outfile << "!" << std::fixed
            << std::setw(9) << "no"
            << std::setw(15) << "x"
            << std::setw(15) << "y"
            << std::setw(15) << "z";
    for (auto &p : res.scalars_at_nodes)
        outfile << std::setw(15) << p.first;
    for (auto &p : res.vectors_at_nodes)
    {
        outfile << std::setw(14) << p.first << "x"
                << std::setw(14) << p.first << "y"
                << std::setw(14) << p.first << "z";
    }
    outfile << std::endl;
    for (auto n : nodes)
    {
        outfile << std::fixed
                << std::setw(10) << n->no
                << std::scientific << std::setprecision(6)
                << std::setw(15) << n->pos(0)
                << std::setw(15) << n->pos(1)
                << std::setw(15) << n->pos(2);
        for (auto &p : res.scalars_at_nodes)
            outfile << std::setw(15) << (*(p.second))[n->row];
        for (auto &p : res.vectors_at_nodes)
            outfile << std::setw(15) << (*(p.second))[n->row](0)
                    << std::setw(15) << (*(p.second))[n->row](1)
                    << std::setw(15) << (*(p.second))[n->row](2);
        outfile << std::endl;
    }
    // Footer
    outfile << std::endl;
    // Close file
    outfile.close();
    std::cout << "done!" << std::endl;
}

void Boundary::write(std::ostream &out) const
{
    out << *groups[0]->tag << " is a Boundary of " << *groups[1]->tag << std::endl;
}