/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wKuttaResidual.h"
#include "wKuttaElement.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"

using namespace tbox;
using namespace flow;

/**
 * @brief Build the residual matrix for a fixed-point iteration, on one Kutta element
 *
 * A = \int psi * v_u*grad_phi dS
 */
Eigen::MatrixXd KuttaResidual::buildFixed(KuttaElement const &ke, std::vector<double> const &phi)
{
    // Get shape functions, Jacobian and Gauss points
    Cache &sCache = ke.surE->getVCache();
    Gauss &sGauss = sCache.getVGauss();
    // Get size
    size_t nRow = ke.nRow;

    // Velocity
    Eigen::RowVectorXd Aup = ke.volE->computeGradient(phi, 0).transpose() * ke.volE->getJinv(0) * ke.volE->getVCache().getDsf(0);

    // Build
    Eigen::MatrixXd A = Eigen::MatrixXd::Zero(nRow, ke.nCol);
    for (size_t k = 0; k < sGauss.getN(); ++k)
    {
        // shape functions
        Eigen::VectorXd const &sf = sCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = sf(ke.teN[i].first);
        // wake contribution
        A += ke.sign * sfp * Aup * sGauss.getW(k) * ke.surE->getDetJ(k);
    }
    return A;
}

/**
 * @brief Build the residual vector, on one Kutta element
 *
 * b = \int psi * (grad_phi)^2 dS
 */
Eigen::VectorXd KuttaResidual::build(KuttaElement const &ke, std::vector<double> const &phi)
{
    // Get shape functions, Jacobian and Gauss points
    Cache &sCache = ke.surE->getVCache();
    Gauss &sGauss = sCache.getVGauss();
    // Get size
    size_t nRow = ke.nRow;

    // Velocity
    Eigen::VectorXd dPhi = ke.volE->computeGradient(phi, 0);
    double dPhi2 = dPhi.squaredNorm();

    // Build
    Eigen::VectorXd b = Eigen::VectorXd::Zero(nRow);
    for (size_t k = 0; k < sGauss.getN(); ++k)
    {
        // shape functions
        Eigen::VectorXd const &sf = sCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = sf(ke.teN[i].first);
        // wake contribution
        b += ke.sign * sfp * dPhi2 * sGauss.getW(k) * ke.surE->getDetJ(k);
    }
    return b;
}

/**
 * @brief Build the gradient of the residual vector with respect to the flow variable (jacobian), on one Kutta element
 *
 * A = \int psi * 2*(grad_phi_u)*dgrad_phi_u dS
 */
Eigen::MatrixXd KuttaResidual::buildGradientFlow(KuttaElement const &ke, std::vector<double> const &phi)
{
    // Build
    Eigen::MatrixXd A = KuttaResidual::buildFixed(ke, phi);
    return 2 * A;
}

/**
 * @brief Build the gradient of the residual vector with respect to the nodes, on one Kutta element
 *
 * A = d( \int psi * (grad_phi)^2 dS )
 *   = \int psi * d(grad_phi)^2 dS
 *   + \int psi * (grad_phi)^2 ddS
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd> KuttaResidual::buildGradientMesh(KuttaElement const &ke, std::vector<double> const &phi)
{
    // Get shape functions and Gauss points
    Cache &surCache = ke.surE->getVCache();
    Gauss &surGauss = surCache.getVGauss();
    // Get Jacobian
    Eigen::MatrixXd const &iJ = ke.volE->getJinv(0);
    std::vector<Eigen::MatrixXd> const &dJ = ke.volE->getGradJ(0);
    // Get size
    size_t nRow = ke.nRow;
    size_t nDim = ke.nDim;
    size_t nCol = ke.nCol;
    size_t nSur = ke.surE->nodes.size();

    // velocity
    Eigen::VectorXd dPhi = ke.volE->computeGradient(phi, 0);

    // Build
    Eigen::MatrixXd Av = Eigen::MatrixXd::Zero(nRow, nDim * nCol);
    Eigen::MatrixXd As = Eigen::MatrixXd::Zero(nRow, nDim * nSur);
    for (size_t k = 0; k < surGauss.getN(); ++k)
    {
        // weight and Jacobian determinant
        double w = surGauss.getW(k);
        double detJ = ke.surE->getDetJ(k);
        std::vector<double> dDetJ = ke.surE->getGradDetJ(k);
        // shape functions
        Eigen::VectorXd const &sf = surCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = sf(ke.teN[i].first);

        // Volume
        for (size_t i = 0; i < nCol; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                // gradient of velocity jump
                Av.col(idx) += ke.sign * w * sfp * 2 * dPhi.transpose() * (-iJ * dJ[idx] * dPhi) * detJ; // psi * dgrad_phi^2
            }
        }
        // Surface
        for (size_t i = 0; i < nSur; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                // gradient of jacobian determinant
                As.col(idx) += ke.sign * w * sfp * dPhi.dot(dPhi) * dDetJ[idx]; // psi * grad_phi^2 * ddetJ
            }
        }
    }
    return std::make_tuple(Av, As);
}
