#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute lifting (linear or nonlinear) potential flow around a NACA 0012 rectangular wing
# Adrien Crovato
#
# Test the nonlinear shock-capturing capability and the Kutta condition
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import math
import flow.utils as floU
import flow.default as floD
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 3*math.pi/180
    M_inf = 0.3
    dim = 3

    # define dimension and mesh size
    spn = 1.0 # wing span
    lgt = 3.0 # channel half length
    hgt = 3.0 # channel half height
    wdt = 3.0 # channel width
    S_ref = 1.*spn # reference area
    c_ref = 1 # reference chord
    fms = 1.0 # farfield mesh size
    nms = 0.02 # nearfield mesh size

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'spn' : spn, 'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msLe' : nms, 'msTe' : nms, 'msF' : fms}
    msh, gmshWriter = floD.mesh(dim, 'models/n0012_3.geo', pars, ['field', 'wing', 'symmetry', 'downstream'], wktp = 'wakeTip')
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, _ = floD.problem(msh, dim, alpha, 0., M_inf, S_ref, c_ref, 0., 0., 0., 'wing', tp = 'teTip')
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    solver = floD.newton(pbl)
    solver.run()
    solver.save(gmshWriter)
    tms['solver'].stop()

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # visualize solution
    floD.initViewer(pbl)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if alpha == 3*math.pi/180 and M_inf == 0.3 and spn == 1.0:
        tests.add(CTest('iteration count', solver.nIt, 3, 1, forceabs=True))
        tests.add(CTest('CL', solver.Cl, 0.135, 5e-2))
        tests.add(CTest('CD', solver.Cd, 0.0062, 1e-2)) # Tranair (NF=0.0062, FF=0.0030), Panair 0.0035
        tests.add(CTest('CS', solver.Cs, 0.012, 1e-2))
        tests.add(CTest('CM', solver.Cm, -0.0278, 1e-2))
    else:
        raise Exception('Test not defined for this flow')
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
