#ifndef KATOPTRON_NEUMANNLIST_H
#define KATOPTRON_NEUMANNLIST_H

#include "katoptron.h"

#include "Map.h"
#include "wNeumann.h"
#include "wTag.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

#include "EnsembleTraits.h"

namespace katoptron
{

/**
 * @brief Class which is used to store the list of the Neumann boundary conditions (BC).
 */
template <typename scalar>
class NeumannList
{
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;

private:
    Kokkos::View<int **, Kokkos::LayoutRight> neumann_elementsList;
    Kokkos::View<scalar ***, Kokkos::LayoutRight> neumann_values_list;

    local_ordinal_type neumann_number;
    local_ordinal_type neumann_size;
    local_ordinal_type neumann_elements;

    local_ordinal_type neumann_time;
    local_ordinal_type neumann_dofs;
    local_ordinal_type neumann_values;

public:
    NeumannList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList);

    /**
     * @brief Return the number of Neumann BC.
     */
    inline local_ordinal_type getNeumannNumber() { return neumann_number; }

    /**
     * @brief Return the number of elements for a
     * given Neumann BC.
     * 
     * Argument:
     * <ul>
     *  <li> i: the BC ID.
     * </ul>
     */
    inline local_ordinal_type getNeumannSize(local_ordinal_type i)
    {
        return neumann_elementsList(i, neumann_size);
    }

    /**
     * @brief Return the local ID of the element e1 of 
     * a given BC.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the BC ID,
     *  <li> e1: the ID of the element on the current BC.
     * </ul>
     */
    inline local_ordinal_type getNeumannElement(local_ordinal_type i, local_ordinal_type e1)
    {
        return neumann_elementsList(i, neumann_elements + e1);
    }

    /**
     * @brief Return whether there is a surface
     * load for the given BC for a given DOF index.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the BC ID,
     *  <li> j: the index of the DOF (0-3 if thermomechanical problems).
     * </ul> 
     */
    inline scalar getNeumannDof(local_ordinal_type i, local_ordinal_type j)
    {
        return neumann_values_list(i, 0, neumann_dofs + j);
    }

    /**
     * @brief Return the value of the surface
     * load for the given BC for a given DOF index.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the BC ID,
     *  <li> j: the index of the DOF (0-3 if thermomechanical problems).
     * </ul> 
     */
    inline scalar getNeumannValue(local_ordinal_type i, local_ordinal_type j)
    {
        return neumann_values_list(i, 0, neumann_values + j);
    }
};

/**
 * @brief NeumannList constructor
 * 
 * This constructor allocates 2 Kokkos::View which store all the required information related to each Neumann BC, including:
 * <ul>
 *  <li> all the Neumann elements,
 *  <li> all the values of the Neumann BC.
 * </ul>
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> map: an RCP to a Map object,
 *  <li> elementsList: an RCP to an ElementList object.
 * </ul>
 */
template <typename scalar>
NeumannList<scalar>::NeumannList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList)
{
    typedef local_ordinal_type local_ordinal_type;
    typedef global_ordinal_type global_ordinal_type;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    local_ordinal_type numDPN = map->numPrimalDPN;

    neumann_number = pbl.nBCs.size();
    local_ordinal_type maxNumElemsPernBC = 0;

    std::vector<std::vector<global_ordinal_type>> nbcElems = {};

    for (auto i = 0; i < neumann_number; ++i)
    {
        std::vector<global_ordinal_type> mynBcElems = {};
        for (auto j = 0; j < pbl.nBCs[i]->tag->elems.size(); ++j)
        {
            local_ordinal_type e = map->mapElems->getLocalElement(pbl.nBCs[i]->tag->elems[j]->no - 1);

            if (elementsList->getElementType(e) == static_cast<int>(tbox::ELTYPE::QUAD4))
                ;
            else if (elementsList->getElementType(e) == static_cast<int>(tbox::ELTYPE::TRI3))
                ;
            else
                continue;

            mynBcElems.push_back(e);
        }
        nbcElems.push_back(mynBcElems);
        if (maxNumElemsPernBC < mynBcElems.size())
            maxNumElemsPernBC = mynBcElems.size();
    }

    neumann_size = 0;
    neumann_elements = 1;
    neumann_elementsList = Kokkos::View<int **, Kokkos::LayoutRight>("R", neumann_number, neumann_elements + maxNumElemsPernBC);

    for (auto i = 0; i < neumann_number; ++i)
    {
        neumann_elementsList(i, neumann_size) = nbcElems[i].size();

        for (auto j = 0; j < nbcElems[i].size(); ++j)
            neumann_elementsList(i, neumann_elements + j) = nbcElems[i][j];
    }

    neumann_time = 0;
    neumann_dofs = 1;
    neumann_values = 1 + numDPN;
    neumann_values_list = Kokkos::View<scalar ***, Kokkos::LayoutRight>("R", neumann_number, 1, 1 + 2 * numDPN);
    for (auto i = 0; i < neumann_number; ++i)
    {
        neumann_values_list(i, 0, neumann_time) = 0.; //to change latter
        for (auto j = 0; j < numDPN; ++j)
            if (pbl.nBCs[i]->which_dof[j])
                neumann_values_list(i, 0, neumann_dofs + j) = 1;
            else
                neumann_values_list(i, 0, neumann_dofs + j) = 0;

        for (auto j = 0; j < numDPN; ++j)
            for (int s = 0; s < ensemble_size; ++s)
            {
                ET::coeff(neumann_values_list(i, 0, neumann_values + j), s) = pbl.nBCs[i]->values[j * ensemble_size + s];
            }
    }
}
}; // namespace katoptron

#endif //KATOPTRON_NEUMANNLIST_H
