#!/usr/bin/env python3
# -*- coding: utf-8; -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from scipy.stats import gamma
from scipy.stats import norm

def grid(M):

    M = M.astype(int)
    TMP = M.shape

    d = TMP[0]

    m_max = 0
    for i in range(0,d):
        if m_max < M[i]:
            m_max = M[i]

    X = np.zeros((m_max,d))
    W = np.zeros((m_max,d))

    for i in range(0,d):
        x, w = np.polynomial.hermite.hermgauss(M[i])
        for j in range(0,M[i]):
            X[j,i] = np.sqrt(2.)*x[j]
            W[j,i] = w[j]
    return X, W

def rho_gamma(g,m_gamma,sigmaS_gamma):
    k = m_gamma**2 / sigmaS_gamma
    theta = sigmaS_gamma / m_gamma

    out_shape = np.shape(g)

    out = np.zeros(out_shape)

    loc = 0. 
    
    length = len(g)

    for i in range(0,length):
        out[i] = gamma.pdf(g[i], k, loc, theta)

    return out

def c_gamma(g,m_gamma,sigmaS_gamma):
    k = m_gamma**2 / sigmaS_gamma
    theta = sigmaS_gamma / m_gamma

    out_shape = np.shape(g)

    out = np.zeros(out_shape)

    loc = 0. 
    
    length = len(g)

    for i in range(0,length):
        out[i] = gamma.cdf(g[i], k, loc, theta)

    return out

def c_inv_gamma(g,m_gamma,sigmaS_gamma):
    k = m_gamma**2 / sigmaS_gamma
    theta = sigmaS_gamma / m_gamma

    out_shape = np.shape(g)

    out = np.zeros(out_shape)

    loc = 0. 
    
    length = len(g)

    for i in range(0,length):
        out[i] = gamma.isf(1.-g[i], k, loc, theta)

    return out

def c_xi(g):

    out = norm.cdf(g)

    return out

def c_inv_xi(g):
    out = norm.isf(1.-g)

    return out

def sigma(c1,c2,rho):

    c = 1. / (np.sqrt(1-rho**2))

    tmp1 = c_inv_xi(c1)
    tmp2 = c_inv_xi(c2)

    if tmp1 == float('Inf'):
        out = 0.
    elif tmp2 == float('Inf'):
        out = 0.
    elif (rho*tmp1)**2 - (2*rho*tmp1*tmp2)  + (rho*tmp2)**2  == float('Inf'):
        out = 0.
    else:
        out = c*np.exp((-((rho*tmp1)**2 - (2*rho*tmp1*tmp2)  + (rho*tmp2)**2) / (2.*(1.-rho**2))))

    return out

def rho_gamma_gamma(g1,m_gamma1,sigmaS_gamma1,g2,m_gamma2,sigmaS_gamma2,rho):

    length1 = len(g1)
    length2 = len(g2)

    out = np.zeros((length1,length2))


    rho_gamma1 = rho_gamma(g1,m_gamma1,sigmaS_gamma1)
    rho_gamma2 = rho_gamma(g2,m_gamma2,sigmaS_gamma2)

    c_gamma1 = c_gamma(g1,m_gamma1,sigmaS_gamma1)
    c_gamma2 = c_gamma(g2,m_gamma2,sigmaS_gamma2)
    
    for i in range(0,length1):
        for j in range(0,length2):
            out[i,j] = rho_gamma1[i]*rho_gamma2[j]*sigma(c_gamma1[i],c_gamma2[j],rho)

    return out


def f_mapping(xi,d,means,variances,P,method):
    out =  np.zeros((d,1))
    tmp =  np.zeros((d,1))
    if method == 'Cholesky':
        L = np.linalg.cholesky(P)
        #print P
        #print L
        #print np.dot(L,L.T)
        tmp = np.dot(L,xi)   
    elif method == 'PCA':
        V = np.diag(variances)
        V_sr = np.sqrt(V)
        Sigma = np.dot(V_sr,np.dot(P,V_sr))
        invV_sr = np.linalg.inv(V_sr)
        eig_val, U = np.linalg.eig(Sigma)
        Lambda = np.diag(eig_val)
        Lambda_sr = np.sqrt(Lambda)
        tmp = np.dot(np.dot(invV_sr,np.dot(U,Lambda_sr)),xi)
    for i in range(0,d):
        out[i] = c_inv_gamma(np.array([c_xi(tmp[i])]),means[i],variances[i])
    
    return out

'''
def f_mapping_old(xi1,m_gamma1,sigmaS_gamma1,xi2,m_gamma2,sigmaS_gamma2,rho):

    out =  np.zeros((2,1))
    out[0] = c_inv_gamma(np.array([c_xi(xi1)]),m_gamma1,sigmaS_gamma1)
    out[1] = c_inv_gamma(np.array([c_xi(rho*xi1+np.sqrt(1-rho**2)*xi2)]),m_gamma2,sigmaS_gamma2)

    return out


def f_mapping_PCA(xi1,m_gamma1,sigmaS_gamma1,xi2,m_gamma2,sigmaS_gamma2,rho):

    Sigma = np.array([[sigmaS_gamma1,rho*np.sqrt(sigmaS_gamma1)*np.sqrt(sigmaS_gamma2)],[rho*np.sqrt(sigmaS_gamma1)*np.sqrt(sigmaS_gamma2),sigmaS_gamma2]])
    invV_sr = np.linalg.inv(np.sqrt(np.diag([sigmaS_gamma1,sigmaS_gamma2])))
    eig_val, U = np.linalg.eig(Sigma)
    Lambda = np.diag(eig_val)
    Lambda_sr = np.sqrt(Lambda)
    xi_array = np.array([[xi1],[xi2]])
    tmp = np.dot(np.dot(invV_sr,np.dot(U,Lambda_sr)),xi_array)
    out =  np.zeros((2,1))
    out[0] = c_inv_gamma(np.array([c_xi(xi1)]),m_gamma1,sigmaS_gamma1)
    out[1] = c_inv_gamma(np.array([c_xi(rho*xi1+np.sqrt(1-rho**2)*xi2)]),m_gamma2,sigmaS_gamma2)

    return out
'''
