/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WELEMENT_H
#define WELEMENT_H

#include "tbox.h"
#include "wObject.h"
#include <vector>
#include <Eigen/Dense>

namespace tbox
{

/**
 * @brief element types (from gmsh)
 * @authors Romain Boman
 */
enum class ELTYPE
{
    UNKNOWN = 0,
    LINE2 = 1,
    TRI3 = 2,
    QUAD4 = 3,
    TETRA4 = 4,
    HEX8 = 5,
    POINT1 = 15
};
#ifndef SWIG
TBOX_API std::ostream &operator<<(std::ostream &out, ELTYPE const &obj);
#endif

/**
 * @brief Base finite element class
 * @authors Romain Boman, Kim Liegeois, Adrien Crovato
 */
class TBOX_API Element : public fwk::wObject
{
    friend class MshCrack; // so that MshCrack can access normal and cg before update
protected:
    int order; ///< order of integration (Gauss quadrature)
    // Flags
    bool hasVals;  ///< true if (precomputed) values have been initialized
    bool hasGrads; ///< true if (precomputed) gradients have been initialized
    bool isVol;    ///< true if element is a volume element
    // Jacobian
    std::vector<double> detJs;                        ///< determinant of Jacobian matrix
    std::vector<Eigen::MatrixXd> Js;                  ///< Jacobian matrix
    std::vector<Eigen::MatrixXd> iJs;                 ///< inverse of Jacobian matrix
    std::vector<std::vector<double>> gradDetJs;       ///< gradient of Jacobian determinant wrt to nodes position
    std::vector<std::vector<Eigen::MatrixXd>> gradJs; ///< gradient of Jacobian matrix wrt to nodes position
    // Volume and CG
    double vol;                  ///< element volume
    Eigen::Vector3d cg;          ///< element center of gravity
    std::vector<double> gradVol; ///< gradient of volume wrt to nodes position
    // Normal vectors
    Eigen::Vector3d normal;                  ///< unit normal vector
    std::vector<Eigen::Vector3d> gradNormal; ///< graident of unit normal vector wrt to nodes position

public:
    int no;                    ///< label
    Tag *ptag;                 ///< physical tag   (group)
    Tag *etag;                 ///< elementary tag (geometry)
    std::vector<Tag *> parts;  ///< partitions
    std::vector<Node *> nodes; ///< nodes

#ifndef SWIG
    Element(int n, Tag *_ptag, Tag *_etag, std::vector<Tag *> &_parts, std::vector<Node *> &nods);
    virtual ~Element() {}
#endif
    virtual ELTYPE type() const
    {
        return ELTYPE::UNKNOWN;
    }

#ifndef SWIG
protected:
    // Jacobian
    void checkJac(size_t k) const;
    virtual void buildJ();
    virtual void buildSurfaceJ();
    virtual void buildGradientJ();
    virtual void buildGradientSurfaceJ();
    // CG and Volume
    void computeCg();
    virtual void computeV();
    virtual void buildGradientV(int dim);
    // Tangent and normal vectors
    virtual std::vector<Eigen::Vector3d> computeTangents() const;
    virtual std::vector<Eigen::Vector3d> buildTangents(size_t k) const;
    virtual void computeNormal();
    virtual void computeGradientNormal();

public:
    // Initialize and update precomputed values
    virtual void initValues(bool isvol, int ordr = 2);
    virtual void initGradients();
    virtual void update();
    // Getters
    virtual Cache &getVCache() const = 0;
    inline bool hasValues() const;
    inline bool hasGradients() const;
    inline double getDetJ(size_t k) const;
    inline Eigen::MatrixXd const &getJ(size_t k) const;
    inline Eigen::MatrixXd const &getJinv(size_t k) const;
    inline std::vector<double> const &getGradDetJ(size_t k) const;
    inline std::vector<Eigen::MatrixXd> const &getGradJ(size_t k) const;
    inline double getVol() const;
    inline Eigen::Vector3d const &getCg() const;
    inline std::vector<double> const &getGradVol() const;
    inline Eigen::Vector3d const &getNormal() const;
    inline std::vector<Eigen::Vector3d> const &getGradNormal() const;
    // Utilities
    virtual Eigen::VectorXd computeGradient(std::vector<double> const &u, size_t k) const;
    Eigen::MatrixXd gp2node(Eigen::MatrixXd &values_at_gp) const;
    virtual void write(std::ostream &out) const override;
#endif
};

#include "wElement.inl"

} // namespace tbox

#endif //WELEMENT_H
