/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wHex8.h"
#include "wSfHex8.h"
#include "wGaussHex8.h"
#include "wCacheHex8.h"
#include "wNode.h"

using namespace tbox;

Hex8::Hex8(int n, Tag *_ptag, Tag *_etag,
           std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
}

/**
 * @brief Return cache handling shape functions and Gauss points for all Hex8 elements
 */
Cache &Hex8::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Initialize precomputed values
 */
void Hex8::initValues(bool isvol, int ordr)
{
    // Set flags
    isVol = isvol;
    order = ordr;
    hasVals = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    Js.resize(ngp);
    iJs.resize(ngp);
    detJs.resize(ngp);
    this->update();
}
/**
 * @brief Initialize precomputed gradients
 */
void Hex8::initGradients()
{
    // Set flags
    hasGrads = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    gradJs.resize(ngp);
    gradDetJs.resize(ngp);
    this->update();
}
/**
 * @brief Update Hex8 precomputed values and gradients
 */
void Hex8::update()
{
    // Update values
    if (hasVals)
    {
        // Jacobian
        if (!isVol)
            throw std::runtime_error("Hex8::update(): the problem must always have the same dimension as Hex8!\n");
        // Hex8 is always of same dimension as Problem (volume element)
        this->buildJ();
        // Volume and CG
        this->computeV();
        this->computeCg();
    }
    // Update gradients
    if (hasGrads)
    {
        // Jacobian
        if (!isVol)
            throw std::runtime_error("Hex8::updateGradients(): the problem must always have the same dimension as Hex8!\n");
        // Hex8 is always of same dimension as Problem (volume element)
        this->buildGradientJ();
        // Volume
        this->buildGradientV(3);
    }
}

/**
 * @brief Compute Jacobian matrix, inverse and determinant
 *
 * J_ij(3,3) = diN_n * xj_n (n is the node)
 */
void Hex8::buildJ()
{
    CacheHex8 &cache = getCache(order);
    for (size_t k = 0; k < cache.getVGauss().getN(); ++k)
    {
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        Eigen::Matrix3d JJ = Eigen::Matrix3d::Zero(); // temporary fixed-size matrix to efficiently compute the inverse
        size_t i = 0;
        for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
            JJ += dff.col(i) * (*it)->pos.transpose();
        Js[k] = JJ;
        iJs[k] = JJ.inverse();
        detJs[k] = JJ.determinant();
        this->checkJac(k);
    }
}

/**
 * @brief Compute the element volume
 */
void Hex8::computeV()
{
    CacheHex8 &cache = getCache(order);
    GaussHex8 &gauss = cache.gauss;

    // Gauss integration
    vol = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        vol += gauss.getW(k) * this->getDetJ(k);
}
