/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTri3.h"
#include "wSfTri3.h"
#include "wGaussTri3.h"
#include "wCacheTri3.h"
#include "wNode.h"

using namespace tbox;

Tri3::Tri3(int n, Tag *_ptag, Tag *_etag,
           std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
}

/**
 * @brief Return cache holding shape functions and Gauss points for all Tri3 elements
 */
Cache &Tri3::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Initialize precomputed values
 */
void Tri3::initValues(bool isvol, int ordr)
{
    // Set flags
    isVol = isvol;
    order = ordr;
    hasVals = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    Js.resize(ngp);
    iJs.resize(ngp);
    detJs.resize(ngp);
    this->update();
}
/**
 * @brief Initialize precomputed gradients
 */
void Tri3::initGradients()
{
    // Set flags
    hasGrads = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    gradJs.resize(ngp);
    gradDetJs.resize(ngp);
    this->update();
}
/**
 * @brief Update precomputed values and gradients
 */
void Tri3::update()
{
    // Update values
    if (hasVals)
    {
        // Jacobian and normal
        // Tri3 is not of same dimension as Problem (surface element)
        if (!isVol)
        {
            this->buildSurfaceJ();
            this->computeNormal();
        }
        // Tri3 is of same dimension as Problem (volume element)
        else
            this->buildJ();
        // Volume and CG
        this->computeV();
        this->computeCg();
    }
    // Update gradients
    if (hasGrads)
    {
        // Jacobian, volume and normal
        // Tri3 is not of same dimension as Problem (surface element)
        if (!isVol)
        {
            this->buildGradientSurfaceJ();
            this->buildGradientV(3);
            this->computeGradientNormal();
        }
        // Tri3 is of same dimension as Problem (volume element)
        else
        {
            this->buildGradientJ();
            this->buildGradientV(2);
        }
    }
}

/**
 * @brief Compute Jacobian determinant
 *
 * detJ = norm(x_i dN_i,xi X x_i dN_i,eta)
 */
void Tri3::buildSurfaceJ()
{
    for (size_t k = 0; k < getCache(order).getVGauss().getN(); ++k)
    {
        std::vector<Eigen::Vector3d> ts = this->buildTangents(k);
        detJs[k] = ts[0].cross(ts[1]).norm();
    }
}
/**
 * @brief Compute Jacobian matrix, inverse and determinant
 *
 * J_ij(2,2) = diN_n * xj_n (n is the node)
 */
void Tri3::buildJ()
{
    CacheTri3 &cache = getCache(order);
    for (size_t k = 0; k < cache.getVGauss().getN(); ++k)
    {
        Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

        Eigen::Matrix2d JJ = Eigen::Matrix2d::Zero(); // temporary fixed-size matrix to efficiently compute the inverse
        size_t i = 0;
        for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
            JJ += dff.col(i) * (*it)->pos.transpose().block<1, 2>(0, 0);
        Js[k] = JJ;
        iJs[k] = JJ.inverse();
        detJs[k] = JJ.determinant();
        this->checkJac(k);
    }
}

/**
 * @brief Compute Jacobian determinant gradient with respect to each node coordinate
 *
 * dDetJ[l = i * 3 + j] = detJ_l * (dt0 \cross t1 + t0 \cross dt1)_l;
 */
void Tri3::buildGradientSurfaceJ()
{
    CacheTri3 &cache = getCache(order);
    for (size_t k = 0; k < getCache(order).getVGauss().getN(); ++k)
    {
        // Shape functions, tangent vectors and determinant
        Eigen::MatrixXd const &dsf = cache.getDsf(k);
        std::vector<Eigen::Vector3d> ts = this->buildTangents(k);
        Eigen::Vector3d detJ = ts[0].cross(ts[1]).normalized();

        // Jacobian gradient
        gradDetJs[k].resize(3 * 3); // 3 nodes and 3 dimensions (x, y, z)
        for (size_t i = 0; i < 3; ++i)
        {
            for (size_t j = 0; j < 3; ++j)
            {
                Eigen::Vector3d dt0 = Eigen::Vector3d::Zero(), dt1 = Eigen::Vector3d::Zero();
                dt0(j) = dsf(0, i);
                dt1(j) = dsf(1, i); // retain sf gradients associated to ith node for dimension j
                gradDetJs[k][i * 3 + j] = detJ.dot(dt0.cross(ts[1]) + ts[0].cross(dt1));
            }
        }
    }
}
/**
 * @brief Compute Jacobian matrix gradient with respect to each node coordinate
 *
 * dJ[l = i * 2 + j](2,2) = dN_i(2,1) * dxy_i(1,2)/dx_l (for node i and dimension j)
 */
void Tri3::buildGradientJ()
{
    CacheTri3 &cache = getCache(order);
    for (size_t k = 0; k < cache.getVGauss().getN(); ++k)
    {
        Eigen::MatrixXd const &dsf = cache.getDsf(k);

        // Jacobian
        gradJs[k].resize(3 * 2); // 3 nodes and 2 dimensions (x, y)
        for (size_t i = 0; i < 3; ++i)
        {
            for (size_t j = 0; j < 2; ++j)
            {
                gradJs[k][i * 2 + j] = Eigen::Matrix2d::Zero();
                gradJs[k][i * 2 + j].col(j) = dsf.col(i); // retain sf gradients associated to ith node for dimension j
            }
        }
        // Determinant
        double detJ = this->getDetJ(k);
        Eigen::MatrixXd const &iJ = this->getJinv(k);
        gradDetJs[k].resize(3 * 2);
        for (size_t i = 0; i < 3 * 2; ++i)
            gradDetJs[k][i] = detJ * (iJ * gradJs[k][i]).trace();
    }
}

/**
 * @brief Return the element surface
 */
void Tri3::computeV()
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;

    // Gauss integration
    vol = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        vol += gauss.getW(k) * this->getDetJ(k);
}
/**
 * @brief Compute volume gradient with respect to each node coordinate
 *
 * dV[i] = w_k * dDetJ[i]_k
 */
void Tri3::buildGradientV(int dim)
{
    GaussTri3 &gauss = getCache(order).gauss;

    // Gradient of volume
    gradVol.resize(3 * dim, 0.); // 3 nodes
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        double w = gauss.getW(k);                         // GP weight
        std::vector<double> dDetJ = this->getGradDetJ(k); // gradient of Jacobian determinant
        for (size_t i = 0; i < dDetJ.size(); ++i)
            gradVol[i] += w * dDetJ[i];
    }
}

/**
 * @brief Compute element tangent vectors
 */
std::vector<Eigen::Vector3d> Tri3::computeTangents() const
{
    Eigen::Vector3d x1 = nodes[0]->pos;
    Eigen::Vector3d x2 = nodes[1]->pos;
    Eigen::Vector3d x3 = nodes[2]->pos;

    return std::vector<Eigen::Vector3d>{(x2 - x1), (x3 - x1)};
}

/**
 * @brief Build element tangent vectors using shape functions at Gauss point k
 */
std::vector<Eigen::Vector3d> Tri3::buildTangents(size_t k) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    std::vector<Eigen::Vector3d> t(2, Eigen::Vector3d::Zero());
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        t[0] += (*it)->pos * dff(0, i);
        t[1] += (*it)->pos * dff(1, i);
    }
    return t;
}

/**
 * @brief Compute element unit normal vector
 */
void Tri3::computeNormal()
{
    std::vector<Eigen::Vector3d> ts = this->computeTangents();
    normal = ts[0].cross(ts[1]).normalized();
}

/**
 * @brief Compute gradient of element unit normal vector with respect to each node coordinate
 */
void Tri3::computeGradientNormal()
{
    // Compute normal and norm
    std::vector<Eigen::Vector3d> ts = this->computeTangents();
    Eigen::Vector3d n = ts[0].cross(ts[1]);
    double nn = n.norm();

    // Compute gradients
    Eigen::Matrix3d inn = (Eigen::Matrix3d::Identity() - n * n.transpose() / (nn * nn)) / nn; // gradient of n/|n|
    gradNormal.resize(3 * 3);                                                                 // 3 nodes and 3 dimensions
    for (size_t i = 0; i < 3; ++i)
    {
        for (size_t j = 0; j < 3; ++j)
        {
            Eigen::Matrix3d dx = Eigen::Matrix3d::Zero();
            dx(j, i) = 1;                                                                                                // dpos_i / dx_j
            gradNormal[i * 3 + j] = inn * ((dx.col(1) - dx.col(0)).cross(ts[1]) + (ts[0]).cross(dx.col(2) - dx.col(0))); // outer gradient times gradient of cross product
        }
    }
}

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 */
Eigen::VectorXd Tri3::computeGradient(std::vector<double> const &u, size_t k) const
{
    // Solution vector
    Eigen::Vector3d ue;
    for (size_t i = 0; i < nodes.size(); ++i)
        ue(i) = u[nodes[i]->row];
    // Gradient in global axis: inv(J)_ij * d_jN_i * ue_i
    return this->getJinv(k) * getCache(order).getDsf(k) * ue;
}
